import time
import logging
from ndn.svs import SVSync
from ndn.storage import Storage
from ndn.encoding import Name, Component
from ndn_hydra.repo.modules.global_view import GlobalView
from ndn_hydra.repo.group_messages.remove import RemoveMessageTlv
from ndn_hydra.repo.group_messages.message import Message, MessageTypes
from ndn_hydra.repo.modules.file_remover import remove_file


def collect_db_garbage(global_view: GlobalView, data_storage: Storage, svs: SVSync, config: dict, logger: logging.Logger) -> None:
    """
    Removes files that have not been accessed in the last month from a node's databases.
    """
    logger.info("\nGARBAGE COLLECTOR: Collecting DB garbage...")
    
    # Remove files that have expired (as based on the expiration_time configuration) 
    for file in global_view.get_files():
        current_time = time.time()
        expire_time = int(file['expiration_time'])
        # If expire_time is 0, file is set to not expire
        if current_time >= expire_time and expire_time != 0:
            # Delete from global view
            global_view.delete_file(file['file_name'])
            logger.info(f"GARBAGE COLLECTOR: Removed {file['file_name']} from global view.")
            # Remove from data_storage from this node if present
            if config['node_name'] in file['stores']:
                remove_file(data_storage, file, config)
                logger.info(f"GARBAGE COLLECTOR: Removed {file['file_name']} from data storage.")

    logger.info("\nGARBAGE COLLECTOR: Finished collecting DB garbage.")


# -------------------------------------------------------------
# NDN Hydra Heartbeat Group Message
# -------------------------------------------------------------
#  @Project: NDN Hydra
#  @Authors: Please check AUTHORS.rst
#  @Source-Code:   https://github.com/tntech-ngin/ndn-hydra
#  @Documentation: https://ndn-hydra.readthedocs.io
#  @Pip-Library:   https://pypi.org/project/ndn-hydra
# -------------------------------------------------------------

from typing import Callable
from ndn.encoding import *
import json
import struct
from ndn_hydra.repo.modules.global_view import GlobalView
from ndn_hydra.repo.modules.favor_calculator import FavorCalculator, FavorParameters, FavorWeights
from ndn_hydra.repo.group_messages.specific_message import SpecificMessage


class HeartbeatMessageTypes:
    NODE_NAME = 84
    FAVOR_PARAMETERS = 85
    FAVOR_WEIGHTS = 86


class HeartbeatMessageTlv(TlvModel):
    node_name = BytesField(HeartbeatMessageTypes.NODE_NAME)
    favor_parameters = ModelField(HeartbeatMessageTypes.FAVOR_PARAMETERS, FavorParameters)
    favor_weights = ModelField(HeartbeatMessageTypes.FAVOR_WEIGHTS, FavorWeights)


class HeartbeatMessage(SpecificMessage):
    def __init__(self, nid: str, seqno: int, raw_bytes: bytes):
        super(HeartbeatMessage, self).__init__(nid, seqno)
        self.message = HeartbeatMessageTlv.parse(raw_bytes)
        self.message.favor_weights = self.decode_favor_weights(self.message.favor_weights)
        self.message.favor_parameters = self.decode_favor_parameters(self.message.favor_parameters)

    @staticmethod
    def decode_favor_weights(favor_weights):
        decoded_weights = {
            'remaining_storage': float(bytes(favor_weights.remaining_storage)),
            'bandwidth': float(bytes(favor_weights.bandwidth)),
            'rw_speed': float(bytes(favor_weights.rw_speed))
        }
        return decoded_weights

    @staticmethod
    def decode_favor_parameters(favor_parameters):
        decoded_params = {
            'rtt': float(bytes(favor_parameters.rtt)),
            'num_users': float(bytes(favor_parameters.num_users)),
            'bandwidth': float(bytes(favor_parameters.bandwidth)),
            'network_cost': float(bytes(favor_parameters.network_cost)),
            'storage_cost': float(bytes(favor_parameters.storage_cost)),
            'remaining_storage': float(bytes(favor_parameters.remaining_storage)),
            'rw_speed': float(bytes(favor_parameters.rw_speed))
        }
        return decoded_params

    async def apply(self, global_view, data_storage, fetch_file, svs, config):
        node_name = self.message.node_name.tobytes().decode()
        favor_parameters = self.message.favor_parameters
        favor_weights = self.message.favor_weights

        favor_calculator = FavorCalculator()
        favor = favor_calculator.calculate_favor(favor_parameters, favor_weights)

        self.logger.debug(f"\n[MSG][HB]   "
                          f"\n\tNode Name={node_name};"
                          f"\n\tfavor={favor}")
        global_view.update_node(node_name, favor, self.seqno)



import sqlite3
from typing import List

from ndn.storage import SqliteStorage
from ndn.encoding import Name, NonStrictName, parse_data

# DataStorage adds additional methods to the SqliteStorage
class DataStorage(SqliteStorage):
    def __init__(self, db_path: str, write_period: int = 10, initialize: bool = True):
        super().__init__(db_path, write_period, initialize)
        self.batch_size = 900
        # self.conn.execute('PRAGMA journal_mode=wal')  # Enable WAL mode

    def put_packet(self, name: NonStrictName, data: bytes) -> None:
        if not self.initialized:
            raise self.UninitializedError("The storage is not initialized.")
        _, _, content, _ = parse_data(data)
        super().put_packet(name, content)

    def remove_packets(self, names: List[NonStrictName]) -> int:
        if not self.initialized:
            raise self.UninitializedError('The storage is not initialized.')

        keys = [self._get_name_bytes_wo_tl(Name.normalize(name)) for name in names]
        if not keys:
            return 0

        total_deleted = 0
        cursor = None

        try:
            self.conn.execute('BEGIN TRANSACTION')
            cursor = self.conn.cursor()

            for i in range(0, len(keys), self.batch_size):
                batch_keys = keys[i:i + self.batch_size]
                placeholders = ', '.join('?' * len(batch_keys))
                query = f'DELETE FROM data WHERE key IN ({placeholders})'

                cursor.execute(query, batch_keys)
                total_deleted += cursor.rowcount

            self.conn.commit()

        except sqlite3.Error as e:
            self.conn.rollback()
            print(f'SQLite error: {e}')

        finally:
            if cursor:
                cursor.close()

        try:
            self.conn.execute('VACUUM')

        except sqlite3.Error as e:
            print(f'SQLite error during VACUUM: {e}')

        # Close connection (remove_packets is performed with a separate connection)
        self.conn.close()

        return total_deleted



# -------------------------------------------------------------
# NDN Hydra Concurrent Segment Fetcher
# -------------------------------------------------------------
#  @Project: NDN Hydra
#  @Authors: Please check AUTHORS.rst
#  @Source-Code:   https://github.com/tntech-ngin/ndn-hydra
#  @Documentation: https://ndn-hydra.readthedocs.io
#  @Pip-Library:   https://pypi.org/project/ndn-hydra
# -------------------------------------------------------------
# NOTE: This concurrent_fetcher was originally written by
#       jonnykong@cs.ucla.edu on 2019-10-15 and later modified
#       to meet the demands of Hydra.
# -------------------------------------------------------------

import asyncio as aio
import logging
from ndn.app import NDNApp
from ndn.types import InterestNack, InterestTimeout
from ndn.encoding import Name, NonStrictName, Component
from tqdm.asyncio import tqdm
from typing import Optional


# An async-generator to fetch data packets concurrently.
async def concurrent_fetcher(app: NDNApp, name: NonStrictName, file_name: NonStrictName, start_block_id: int,
                             end_block_id: Optional[int], semaphore: aio.Semaphore, **kwargs):
    cur_id = start_block_id
    final_id = end_block_id if end_block_id is not None else 0x7fffffff
    is_failed = False
    tasks = []
    recv_window = cur_id - 1
    seq_to_data_packet = dict()  # Buffer for out-of-order delivery
    received_or_fail = aio.Event()

    # Progress bar
    total_blocks = final_id - start_block_id + 1
    progress_bar = tqdm(total=total_blocks, desc='Fetching data', unit='block')

    async def _retry(seq: int):
        """
        Retry 3 times fetching data of the given sequence number or fail.
        :param seq: block_id of data
        """
        nonlocal app, name, file_name, semaphore, is_failed, received_or_fail, final_id
        int_name = Name.normalize(name) + [Component.from_segment(seq)]
        # print(Name.to_str(int_name))
        key = Name.normalize(file_name) + [Component.from_segment(seq)]
        # print(Name.to_str(key))

        trial_times = 0
        while True:
            trial_times += 1
            if trial_times > 3:
                semaphore.release()
                is_failed = True
                received_or_fail.set()
                return
            try:
                # logging.info('Express Interest: {}'.format(Name.to_str(int_name)))
                data_name, meta_info, content, data_bytes = await app.express_interest(
                    int_name, need_raw_packet=True, can_be_prefix=False, must_be_fresh=False, lifetime=4000, **kwargs)

                # Save data and update final_id
                # logging.info('Received data: {}'.format(Name.to_str(data_name)))
                seq_to_data_packet[seq] = (data_name, meta_info, content, data_bytes, key)
                if meta_info is not None and meta_info.final_block_id is not None:
                    final_id = Component.to_number(meta_info.final_block_id)
                progress_bar.update()
                break
            except InterestNack as e:
                logging.info(f'\nNacked with reason={e.reason} {Name.to_str(int_name)}')
            except InterestTimeout:
                logging.info(f'\nTimeout {Name.to_str(int_name)}')
        semaphore.release()
        received_or_fail.set()

    async def _dispatch_tasks():
        """
        Dispatch retry() tasks using semaphore.
        """
        nonlocal semaphore, tasks, cur_id, final_id, is_failed
        while cur_id <= final_id:
            await semaphore.acquire()
            if is_failed:
                received_or_fail.set()
                semaphore.release()
                break
            task = aio.get_event_loop().create_task(_retry(cur_id))
            tasks.append(task)
            cur_id += 1

    aio.get_event_loop().create_task(_dispatch_tasks())
    while True:
        await received_or_fail.wait()
        received_or_fail.clear()
        # Re-assemble bytes in order
        while recv_window + 1 in seq_to_data_packet:
            yield seq_to_data_packet[recv_window + 1]
            del seq_to_data_packet[recv_window + 1]
            recv_window += 1
        # Return if all data have been fetched, or the fetching process failed
        if recv_window == final_id:
            await aio.gather(*tasks)
            break
        if is_failed:
            await aio.gather(*tasks)
            # New data may return during gather(), need to check again
            while recv_window + 1 in seq_to_data_packet:
                yield seq_to_data_packet[recv_window + 1]
                del seq_to_data_packet[recv_window + 1]
                recv_window += 1
            break
    progress_bar.close()



#  @Project: NDN Hydra
#  @Date:    2024-03-04
#  @Authors: Please check AUTHORS.rst
#  @Source-Code:   https://github.com/tntech-ngin/ndn-hydra
#  @Documentation: https://ndn-hydra.readthedocs.io
#  @Pip-Library:   https://pypi.org/project/ndn-hydra
# -------------------------------------------------------------
import numpy as np
from ndn.encoding import *
from typing import Union, Dict
import shutil


class FavorParameterTypes:
    RTT = 501
    NUM_USERS = 502
    BANDWIDTH = 503
    NETWORK_COST = 504
    STORAGE_COST = 505
    REMAINING_STORAGE = 506
    RW_SPEED = 507


class FavorWeightsTypes:
    REMAINING_STORAGE = 508
    BANDWIDTH = 509
    RW_SPEED = 510


class FavorWeights(TlvModel):
    remaining_storage = BytesField(FavorWeightsTypes.REMAINING_STORAGE)
    bandwidth = BytesField(FavorWeightsTypes.BANDWIDTH)
    rw_speed = BytesField(FavorWeightsTypes.RW_SPEED)


class FavorParameters(TlvModel):
    rtt = BytesField(FavorParameterTypes.RTT)
    num_users = BytesField(FavorParameterTypes.NUM_USERS)
    bandwidth = BytesField(FavorParameterTypes.BANDWIDTH)
    network_cost = BytesField(FavorParameterTypes.NETWORK_COST)
    storage_cost = BytesField(FavorParameterTypes.STORAGE_COST)
    remaining_storage = BytesField(FavorParameterTypes.REMAINING_STORAGE)
    rw_speed = BytesField(FavorParameterTypes.RW_SPEED)


class FavorCalculator:
    """
    A class for abstracting favor calculations between two nodes.
    """

    @staticmethod
    def calculate_favor(
            favor_parameters: Union[FavorParameters, Dict[str, float]],
            favor_weights: Union[FavorWeights, Dict[str, float]]
    ) -> float:
        favor = (favor_weights['remaining_storage'] * favor_parameters['remaining_storage']
                 + favor_weights['bandwidth'] * favor_parameters['bandwidth']
                 + favor_weights['rw_speed'] * favor_parameters['rw_speed'])
        return int(favor)




import asyncio as aio
from ndn_hydra.repo.modules.data_storage import DataStorage

# Responsible for removing data from the data storage
def remove_file(data_storage, file, config):
    data_storage = DataStorage(config['data_storage_path'])
    keys = [file['file_name'] + f'/seg={seq}' for seq in range(file['packets'])]
    aio.get_event_loop().run_in_executor(None, data_storage.remove_packets, keys)



# -------------------------------------------------------------
# NDN Hydra Fetch Client
# -------------------------------------------------------------
#  @Project: NDN Hydra
#  @Authors: Please check AUTHORS.rst
#  @Source-Code:   https://github.com/tntech-ngin/ndn-hydra
#  @Documentation: https://ndn-hydra.readthedocs.io
#  @Pip-Library:   https://pypi.org/project/ndn-hydra
# -------------------------------------------------------------

from asyncio import Semaphore
import logging
import time
from ndn.app import NDNApp
from ndn.encoding import FormalName, Component, Name, ContentType
import os
from ndn_hydra.repo.utils.concurrent_fetcher import concurrent_fetcher


class HydraFetchClient(object):
    def __init__(self, app: NDNApp, client_prefix: FormalName, repo_prefix: FormalName) -> None:
        """
        This client fetches data packets from the remote repo.
        :param app: NDNApp.
        :param client_prefix: NonStrictName. Routable name to client.
        :param repo_prefix: NonStrictName. Routable name to remote repo.
        """
        self.app = app
        self.client_prefix = client_prefix
        self.repo_prefix = repo_prefix

    async def fetch_file(self, file_name: FormalName, local_filename: str = None, overwrite: bool = False) -> None:
        """
        Fetch a file from remote repo, and write to the current working directory.
        :param name_at_repo: NonStrictName. The name with which this file is stored in the repo.
        :param local_filename: str. The filename of the retrieved file on the local file system.
        :param overwrite: If true, existing files are replaced.
        """
        name_at_repo = self.repo_prefix + file_name + [Component.from_segment(0)]

        # If no local filename is provided, store file with last name component
        # of repo filename
        if local_filename is None:
            local_filename = Name.to_str(file_name)
            local_filename = os.path.basename(local_filename)

        # If the file already exists locally and overwrite=False, retrieving the file makes no
        # sense.
        if os.path.isfile(local_filename) and not overwrite:
            raise FileExistsError("{} already exists".format(local_filename))

        # Ping with only one data packet at the start
        b_array = bytearray()
        start_index = 0
        end_index = None
        data_name, meta_info, content, data_bytes = await self.app.express_interest(
            name_at_repo, need_raw_packet=True, can_be_prefix=False, must_be_fresh=False, lifetime=4000)

        forwarding_hint = []
        name_at_repo = name_at_repo[:-1]

        if meta_info.content_type == ContentType.NACK:
            print("Distributed Repo does not have that file.")
            return

        if meta_info.content_type == ContentType.LINK:
            # name_at_repo = Name.from_str(bytes(content).decode())
            forwarding_hint = [(1, bytes(content).decode())]
        # else:
        #     # name_at_repo = name_at_repo[:-1]
        #     # start_index = start_index + 1
        #     # print(name_at_repo)
        #     b_array.extend(content)

        end_index = Component.to_number(meta_info.final_block_id)

        # print(Name.to_str(data_name))

        # Fetch the rest of the file.
        if start_index <= end_index:
            async for (_, _, content, _, _) in concurrent_fetcher(self.app, name_at_repo, Name.from_str(local_filename), start_index, end_index, Semaphore(10), forwarding_hint=forwarding_hint):
                b_array.extend(content)

        # After b_array is filled, sort out what to do with the data.
        if len(b_array) > 0:
            print(f'Fetching completed, writing to file {local_filename}')

            # Create folder hierarchy
            local_folder = os.path.dirname(local_filename)
            if local_folder:
                os.makedirs(local_folder, exist_ok=True)

            # Write retrieved data to file
            if os.path.isfile(local_filename) and overwrite:
                os.remove(local_filename)
            with open(local_filename, 'wb') as f:
                f.write(b_array)

            return name_at_repo
        else:
            print("Client Fetch Command Failed.")


# -------------------------------------------------------------
# NDN Hydra Remove Group Message
# -------------------------------------------------------------
#  @Project: NDN Hydra
#  @Authors: Please check AUTHORS.rst
#  @Source-Code:   https://github.com/tntech-ngin/ndn-hydra
#  @Documentation: https://ndn-hydra.readthedocs.io
#  @Pip-Library:   https://pypi.org/project/ndn-hydra
# -------------------------------------------------------------

import asyncio as aio
from typing import Callable
from ndn.encoding import *
from ndn_hydra.repo.modules.global_view import GlobalView
from ndn_hydra.repo.group_messages.specific_message import SpecificMessage
from ndn_hydra.repo.modules.file_remover import remove_file


class RemoveMessageTypes:
    NODE_NAME = 84
    FAVOR = 86


class RemoveMessageTlv(TlvModel):
    node_name = BytesField(RemoveMessageTypes.NODE_NAME)
    favor = BytesField(RemoveMessageTypes.FAVOR)
    file_name = NameField()


class RemoveMessage(SpecificMessage):
    def __init__(self, nid: str, seqno: int, raw_bytes: bytes):
        super(RemoveMessage, self).__init__(nid, seqno)
        self.message = RemoveMessageTlv.parse(raw_bytes)

    async def apply(self, global_view, data_storage, fetch_file, svs, config):
        node_name = self.message.node_name.tobytes().decode()
        file_name = Name.to_str(self.message.file_name)

        self.logger.info(f"\n[MSG][REMOVE]   "
                         f"\n\tFile name={file_name}")
        file = global_view.get_file(file_name)
        if not file:
            self.logger.warning('nothing to remove')
        else:
            # Delete from global view
            global_view.delete_file(file_name)
            # Remove from data_storage from this node if present
            if config['node_name'] in file['stores']:
                remove_file(data_storage, file, config)

        global_view.update_node(node_name, float(self.message.favor.tobytes().decode()), self.seqno)
